/*
 * Compaq TC1000 Touchscreen Device
 * Version 0.3
 * Copyright (c) Frank Bruno <fbruno@attbi.com>
 *
 * Can be freely distributed and used under the terms of the GNU GPL.
 * 
 * This module file is a hack. It is the DSDT init routine for the 
 * COM1 interface of the Via vt82c686b chipset. This should really go
 * in the serial.c file somewhere, or ideally the ACPI interpretter
 * Should handle all of this.
 *
 * 2003-09-08  Adam: added MODULE_LICENSE("GPL") so we don't taint the kernel.

 * 2004-06-02  complete rewrite, after it always locked up my TC1000; got the
 *	       spec from VIA and there seemed to be at least one bug.
 *             David Kühling <dvdkhlng@gmx.de>
 *
 *             Frank's address was suspeneded, couldn't contact him, so I just
 *             put my name and email into MODULE_AUTHOR.  Sorry, Frank.

 * 2004-05-21  Tomasz Sterna <tomek@xiaoka.com>
 *             I ported fpi2002-0.2 module to kernel 2.6 using viahss-0.92
 *             as a base and then David's fpi2002-0.4 using the same technique
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/fs.h>       /* The character device 
                             * definitions are here */

static const unsigned short superio_index_port = 0x3F0;
static const unsigned short superio_data_port = 0x3F1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
spinlock_t driver_lock = SPIN_LOCK_UNLOCKED;
#endif

/* before we can read/write Super-I/O chip's configuration at 0x3F0, we first
 * have to set that feature in the 82C686 PCI configuration space */
static void enable_superio_config (struct pci_dev *pcidev, int enable)
{
   unsigned char val;
   pci_read_config_byte (pcidev, 0x85, &val);

   val &= ~(0x02);
   if (enable)
      val |= 0x02;

   pci_write_config_byte (pcidev, 0x85, val);
}

static unsigned char superio_config_read (int index)
{
   outb(index, superio_index_port);
   return inb(superio_data_port);
}

static void superio_config_write (int index, unsigned char data)
{
   outb (index, superio_index_port);
   outb (data, superio_data_port);
}

static int fpi2002_init(void)
{
   struct pci_dev *pcidev = NULL;
   unsigned char val;

   /* See if we have a via chip */
   pcidev = pci_find_device (PCI_VENDOR_ID_VIA, PCI_DEVICE_ID_VIA_82C686, NULL);
   if (pcidev) 
   {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
      spin_lock_irq(&driver_lock);
#else
      cli();
#endif

      /* begin of Super-I/O chip configuration */
      enable_superio_config (pcidev, 1);
     
      /* Super-I/O function select: if serial port was enabled, disable it
       * now */
      val = superio_config_read (0xE2);
      superio_config_write (0xE2, val & (~0x4));

      /* set serial port 1 I/O base address to the usual 0x3F8 */
      superio_config_write (0xE7, 0x3F8 >> 2);

      /* set the IRQ of serial port 1 to IRQ4.  IRQ routing is controlled from
       * the PCI configuration space of the VIA82C686 at offset 0x52 ("PNP IRQ
       * Routing 2"), the lower nibble sets serial port 1 IRQ, the higher
       * nibble set serial port 2 IRQ.  */
      pci_read_config_byte (pcidev, 0x52, &val);
      val = (val&0xF0) | 0x04;
      pci_write_config_byte (pcidev, 0x52, val);

      /* now we can enable serial port */
      val = superio_config_read (0xE2);
      superio_config_write (0xE2, val | 0x04);

      /* end of Super-I/O configuration */
      enable_superio_config (pcidev, 0);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
      spin_unlock_irq(&driver_lock);
#else
      sti();
#endif
      printk (KERN_INFO "Enabled FPI2002 Interface\n");
   }
   else 
   {
      printk (KERN_INFO "Couldn't locate VIA chipset\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
      return -ENODEV;
#else
      return -1;
#endif
   }
   return 0;
}

static void fpi2002_exit(void)
{
   struct pci_dev *pcidev = NULL;
   unsigned char val;

   pcidev = pci_find_device (PCI_VENDOR_ID_VIA,PCI_DEVICE_ID_VIA_82C686,NULL);
   if (pcidev) 
   {

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
      spin_lock_irq(&driver_lock);
#else
      cli();
#endif

      /* begin of Super-I/O chip configuration */
      enable_superio_config (pcidev, 1);

      /* set serial port 1 base address to zero; this should disable the serial
       * port (or make it unaccessible at least) */
      superio_config_write (0xE7, 0x00);

      /* disable interrupt for serial port 1 */
      pci_read_config_byte (pcidev, 0x52, &val);
      val = (val&0xF0);
      pci_write_config_byte (pcidev, 0x52, val);
    
      /* end of Super-I/O config */
      enable_superio_config (pcidev, 0);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
      spin_unlock_irq(&driver_lock);
#else
      sti();
#endif
      printk (KERN_INFO "FPI2002 disabled\n");
   }
}

module_init(fpi2002_init);
module_exit(fpi2002_exit);
MODULE_DESCRIPTION("FPI2002 port enabler");
MODULE_AUTHOR("David Kuehling <dvdkhlng@gmx.de>");

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
